import neko.io.Input;
import neko.io.File;
import neko.Lib;

class Command
{
   public var mPos:Int;
   public var mBSize:Int;
   public var mDiff:Int;
   public var mCode:String;
   public var mArgs:String;
   public var mDebug:String;
   public var mLabel:Int;
   public var mLID:Int;

   public function new(inCode:String,inArgs:String,inDebug:String,inLabel:Int)
   {
      mPos = 0;
      mBSize = 1;
      mCode = inCode;
      mArgs = inArgs;
      mDebug = inDebug;
      mLabel = inLabel;
      mLID = 0;
      mDiff = 0;

      //neko.Lib.print("  // " + mCode + "(" + mArgs + ") = " + mBSize + "\n");
   }

   public function Clear()
   {
      mBSize = 0;
      mCode = "";
   }

   public function GetSize() : Int
   {
    var s = 
      switch(mCode)
      {
         case "findpropstrict" : 2;
         case "findprop" : 2;
         case "getproperty" : 2;
         case "pushdouble" : 2;
         case "pushint" : 2;
         case "astype" : 2;
         case "getlex" : 2;
         case "" : 0;
         default:
            mBSize;
      }

      if (mBSize!=s)
      {
         //neko.Lib.print("  // " + mCode + "(" + mArgs + ")" +
             //" changing size from " + mBSize + " to " + s + "\n");
      }

      return s;
   }
}

typedef CommandList = Array<Command>;

class Traslator
{
   var mCallProp : EReg;
   public var mJumps : Array<Int>;

   public function new()
   {
      mCallProp = new EReg("(\\S+) \\((\\d+)\\)","");
      mJumps = new Array<Int>();
   }

   public function JumpCode(inCondition:String,inID:Int,inDiff:Int)
   {
      // backwards jump...
      if (inDiff<0)
      {
         return "   ]);\n   jump_" + inID + "(" + inCondition +
             ");\n   mCtx.ops([";
      }
      else
      {
         if (mJumps[inID] == null)
            mJumps[inID] = 0;
         var n = mJumps[inID]++;
         return "   ]);\n   var jump_" + inID + "_" + n + 
            " = mCtx.jump(" + inCondition + ");\n   mCtx.ops([";
      }
   }

   public function Translate(inCommand:Command) : String
   {
      var args = inCommand.mArgs;

      if (inCommand.mCode=="callproperty")
      {
         mCallProp.match(args);
         return "OCallProperty(mCtx.property(\"" + mCallProp.matched(1) +
                   "\")," + mCallProp.matched(2) + "),";
      }


      return switch(inCommand.mCode)
      {
         case "findpropstrict" : "OFindPropStrict(mCtx.property(\"" + args +
                   "\")),";
         case "findprop" : "OFindProp(" + args + "),";
         case "getproperty" :
            if (args=="null")
               "OGetProp(mCtx.arrayProp),";
            else
               "OGetProp(mCtx.property(\"" + args + "\")),";
         case "debugline" : "ODebugLine(" + args + "),";
         case "debugfile" : "ODebugFile(mCtx.string(" + args + ")),";
         case "pushscope" : "OScope,";
         case "pushnull" : "ONull,";
         case "pushfalse" : "OFalse,";
         case "pushtrue" : "OTrue,";
         case "pushnan" : "ONaN,";
         case "pushbyte" : "OSmallInt(" + args + "),";
         case "pushdouble" : "OFloat(mCtx.float(" + args + ")),";
         case "pushint" : "OIntRef(mCtx.int(" + args + ")),";
         case "astype" : "OAsType(mCtx.property(\"" + args + "\")),";
         case "label" : "OLabel,";
         case "convert_i" : "OToInt,";
         case "convert_s" : "OToString,";
         case "convert_b" : "OToBool,";
         case "convert_d" : "OToNumber,";
         case "getlocal0" : "OReg(0),";
         case "getlocal1" : "OReg(1),";
         case "getlocal2" : "OReg(2),";
         case "getlocal3" : "OReg(3),";
         case "getlocal4" : "OReg(4),";
         case "getlocal" : "OReg(" + args + "),";
         case "setlocal" : "OSetReg(" + args + "),";
         case "multiply_i" : "OOp(OpIMul),";
         case "multiply" : "OOp(OpMul),";
         case "lessequals" : "OOp(OpLte),";
         case "pop" : "OPop,";
         case "iffalse" : JumpCode("JFalse",inCommand.mLID,inCommand.mDiff);
         case "iftrue" :  JumpCode("JTrue",inCommand.mLID,inCommand.mDiff);
         case "jump" :  JumpCode("JAlways",inCommand.mLID,inCommand.mDiff);
         case "inclocal_i" : "OIncrIReg(" + args + "),";
         case "inclocal" : "OIncrReg(" + args + "),";
         case "declocal_i" : "ODecrIReg(" + args + "),";
         case "declocal" : "ODecrReg(" + args + "),";
         case "add_i" : "OOp(OpIAdd),";
         case "add" : "OOp(OpAdd),";
         case "subtract_i" : "OOp(OpISub),";
         case "subtract" : "OOp(OpSub),";
         case "getlex" : "OGetLex(mCtx.property(\"" + args + "\")),";
         case "returnvalue" : "ORet,";
         case "returnvoid" : "ORetVoid,";
         case "lessthan" : "OOp(OpLt),";
         case "greaterthan" : "OOp(OpGt),";
         case "equals" : "OOp(OpEq),";
         case "not" : "OOp(OpNot),";
         default: "############# Unknown(" + inCommand.mCode + "," + args + ")";
      }
   }

}


class Abc2Hxasm
{
   var mDebugFile : Array<String>;
   var mLabelPos : Array<Int>;
   var mCommands : CommandList;
   var mFunc : String;
   var mArgs : String;
   var mReturn : String;

   var mMaxStack : Int;
   var mLocals : Int;

   var mTranslator : Traslator;

   public function new()
   {
      mTranslator = new Traslator();
   }

   public static function print(inText:String) { neko.Lib.print(inText); }

   function PrintProlog()
   {
   }


   public function Dump(inVerbose:Bool)
   {
      var label = ~/^L(\d+)\s*$/;

      if (inVerbose)
      {
         print("   // create a new bytecode Context\n");
         print("   mCtx = new hxasm.Context();\n");
         print("   // defines a class called Main\n");
         print("   mCtx.beginClass(\"Main\");\n");
      }

      //var tint = mCtx.type("int");
      var arg_types = mArgs.split(",");
      var type_list = "[";
      for(arg in arg_types)
      {
         type_list += "mCtx.type(\"" + arg + "\"),";
      }
      type_list += "]";
      var return_type =  "mCtx.type(\"" + mReturn + "\")";


      if (inVerbose)
      {
         print("   var m = mCtx.beginMethod(\"" + mFunc + "\"," + type_list +
                      "," + return_type + ");\n");
         print("   m.maxStack = " + mMaxStack + ";\n");
     }

     print("   mCtx.ops([\n");

      var pos = 0;
      for( command in mCommands )
      {
         command.mPos = pos;
         pos += command.GetSize();
      }


      for( command in mCommands )
      {
         print(command.mDebug);
         if (command.mCode=="label")
         {
            print("   ]);\n");
            print("   var jump_" + command.mLabel + "=mCtx.backwardJump();\n");
            print("   mCtx.ops([\n");
         }
         else if (command.mLabel>0)
         {
            print("   ]);\n");
            var n = mTranslator.mJumps[command.mLabel];
            for(i in 0...n)
               print("   jump_" + command.mLabel + "_" + i + "();\n");
            print("   mCtx.ops([\n");
         }

         if (command.mCode!="")
         {
            var args = command.mArgs;
            if (label.match(args))
            {
               command.mLID = Std.parseInt(label.matched(1));
               var bpos =
                 mCommands[mLabelPos[command.mLID]].mPos;
               command.mDiff = bpos - command.mPos - 4;
            }
            print("     " + mTranslator.Translate(command) + "\n");
         }
      }

      print("   ]);\n");

      if (inVerbose)
      {

         print("   // we are done with all the bytecode writing\n");
         print("   mCtx.finalize();\n");
 
         print("   // create an output and write the bytecode to it\n");
         print("   var o = new hxasm.Output();\n");
         print("   hxasm.Writer.write(o,mCtx);\n");
         print("   var swf = o.getBytes();\n");
 
         print("   // load the SWF bytes\n");
         print("   var loader = new flash.display.Loader();\n");
         print("   loader.contentLoaderInfo.addEventListener(\n" +
                      "      flash.events.Event.COMPLETE,onLoaded);\n");
         print("   loader.loadBytes(swf);\n");
      }

   }

   public function Optimise()
   {
      var last:Command = null;
      for( command in mCommands )
      {
         if (command.mCode=="debugfile" || command.mCode=="debugline")
            command.Clear();

         if (last!=null)
         {
            if (last.mCode == "astype" && command.mCode=="astype")
            {
               last.Clear();
            }
            else if (last.mCode=="findpropstrict" &&
               command.mCode=="getproperty" && last.mArgs == command.mArgs )
            {
               last.Clear();
               command.mCode = "getlex";
            }
         }

         last = command;
      }
   }


   public function ParseFunction(inDef:String,inFile:neko.io.Input) : Bool
   {
      //var extract = ~/^\s*function (\S+)\(.*\):\(\S+\)/;
      var extract = ~/function (\S+)\((.*)\):(\S+)/;
      if (!extract.match(inDef))
      {
         print("Unkown function definition: " + inDef + "\n");
         return false;
      }
      mFunc = extract.matched(1);
      mArgs = extract.matched(2);
      mReturn = extract.matched(3);
      print(" // Function :" + mFunc + "\n");
      print(" // Args     :" + mArgs + "\n");
      print(" // Returns  :" + mReturn + "\n");

      var start = ~/^\s+{/;
      var line = inFile.readLine();
      if (!start.match(line))
      {
         print("No open { found.");
         return false;
      }

      var info = ~/local_count=(\d+) max_scope=(\d+) max_stack=(\d+) code_len=(\d+)/;

      line = inFile.readLine();
      if (!info.match(line))
      {
         print("No function prolog found.");
         return false;
      }

      mLocals = Std.parseInt( info.matched(1) );
      mMaxStack = Std.parseInt( info.matched(3) );
      //print("Locals   : " + mLocals + "\n");
      // print("Scope    : " + info.matched(2) + "\n");
      //print("Stack    : " + mMaxStack + "\n");
      //print("Length   : " + info.matched(4) + "\n");

      var stop = ~/^\s+}/;
      var blank = ~/^\s*$/;
      var label = ~/^\s+L(\d+):/;
      var pending_label = 0;
      var code = ~/^\s+(\d+)\s+(\S+)\s*(.*)/;
      var comment = ~/(.*)\/\//;

      mCommands = new CommandList();
      mLabelPos = new Array<Int>();
      var debug_line = "";

      var pos = 0;
      while(true)
      {
         line = inFile.readLine();
         if (stop.match(line))
            break;
         if (blank.match(line))
            continue;
         if (label.match(line))
         {
            pending_label = Std.parseInt(label.matched(1));
            continue;
         }

         if (comment.match(line))
            line = comment.matched(1);

         if (!code.match(line))
         {
            print("Unexpected line : " + line + "\n");
            return false;
         }


         var offset = Std.parseInt( code.matched(1) );
         if (pos>0)
            mCommands[ mCommands.length - 1 ].mBSize = offset - pos;
         pos = offset;
         var opcode = code.matched(2);
         var args = code.matched(3);

         if (pending_label>0)
         {
            mLabelPos[pending_label] = mCommands.length;
         }


         if (opcode=="debugfile" || opcode=="debugline")
         {
            if (mDebugFile!=null && opcode=="debugline")
            {
               var line = Std.parseInt(args) - 1;
               debug_line += "   // " + mDebugFile[line] + "\n";
            }
         }

         mCommands.push( new Command(opcode,args,debug_line,pending_label) );

         pending_label = 0;
         debug_line = "";
      }

      return true;
      // print("Processed " + commands.length + " commands.\n" );
   }

   public function Parse(inFilename:String,inFunction:String) : Bool
   {
      var in_file:neko.io.FileInput;
      
      try { in_file = File.read(inFilename,true); }
      catch (e:Dynamic)
      {
         print("Could not open " + inFilename + "\n");
         return false;
      }

      var result = false;
      var func = new EReg("function " + inFunction + "\\(","");
      var line_no = 0;
      try {
         while(true)
         {
            var line = in_file.readLine();
            if (func.match(line))
            {
               result = ParseFunction(line,in_file);
               break;
            }
            line_no++;
         }
      }
      catch (e:Dynamic) { }

      in_file.close();
      return result;
   }


   public function LoadDebug(inFilename:String):Bool
   {
      var in_file:neko.io.FileInput;
      
      try { in_file = File.read(inFilename,true); }
      catch (e:Dynamic)
      {
         print("Could not debug file " + inFilename + "\n");
         return false;
      }

      mDebugFile = new Array<String>();

      try {
         while(true)
            mDebugFile.push( in_file.readLine() );
      }
      catch (e:Dynamic) { }

      in_file.close();
      return true;
   }


   public static function main() :Void
   {
        var args = neko.Sys.args();
        if (args.length!=2 && args.length!=3)
        {
           print("Usage: Abc2Hxasm file.abc function_name [src_file.hx].\n");
           return;
        }

        var app:Abc2Hxasm = new Abc2Hxasm();
        if (args.length==3)
        {
           if (!app.LoadDebug(args[2]))
              return;
        }

        var ok = app.Parse(args[0],args[1]);
        if (ok)
        {
           app.Optimise();
           app.Dump(false);
        }
   }

}

